# Copyright © 2026 Igalia S.L.
# SPDX-License-Identifier: MIT
#
# Generates api_trace.c: per-entrypoint wrappers that log GL calls to stderr
# and forward to the real dispatch. Installed at context create when
# MESA_VERBOSE=api is set.

import apiexec
import gl_XML
import license


TRACE_ARRAY_BUFSZ = 512


header = """
#include <inttypes.h>
#include <stdio.h>

#include "glapi/glapi/glapi.h"
#include "main/api_trace_helpers.h"
#include "main/context.h"
#include "main/enums.h"
#include "main/errors.h"
#include "dispatch.h"
"""


TYPE_FORMAT = {
    'GLenum':      ('%s',          '_mesa_enum_to_string({p})'),
    'GLboolean':   ('%s',          '{p} ? "GL_TRUE" : "GL_FALSE"'),
    'GLbitfield':  ('0x%x',        '{p}'),
    'GLbyte':      ('%d',          '{p}'),
    'GLubyte':     ('%u',          '{p}'),
    'GLshort':     ('%d',          '{p}'),
    'GLushort':    ('%u',          '{p}'),
    'GLint':       ('%d',          '{p}'),
    'GLuint':      ('%u',          '{p}'),
    'GLint64':     ('%" PRId64 "', '(int64_t){p}'),
    'GLuint64':    ('%" PRIu64 "', '(uint64_t){p}'),
    'GLsizei':     ('%d',          '{p}'),
    'GLintptr':    ('%" PRIdPTR "','(intptr_t){p}'),
    'GLsizeiptr':  ('%" PRIdPTR "','(intptr_t){p}'),
    'GLfloat':     ('%f',          '{p}'),
    'GLclampf':    ('%f',          '{p}'),
    'GLdouble':    ('%f',          '{p}'),
    'GLclampd':    ('%f',          '{p}'),
    'GLfixed':     ('%d',          '{p}'),
    'GLclampx':    ('%d',          '{p}'),
    'GLhalfNV':    ('0x%x',        '{p}'),
    'GLvdpauSurfaceNV': ('%" PRIdPTR "', '(intptr_t){p}'),
    'GLsync':      ('%p',          '(void *){p}'),
    'GLhandleARB': ('%u',          '(unsigned int){p}'),
    'GLDEBUGPROC':    ('%p',       '(void *){p}'),
    'GLDEBUGPROCARB': ('%p',       '(void *){p}'),
    'GLDEBUGPROCAMD': ('%p',       '(void *){p}'),
    'GLDEBUGPROCKHR': ('%p',       '(void *){p}'),
    'GLVULKANPROCNV': ('%p',       '(void *){p}'),
}

TYPE_ALIASES = {
    'GLint64EXT':    'GLint64',
    'GLuint64EXT':   'GLuint64',
    'GLintptrARB':   'GLintptr',
    'GLsizeiptrARB': 'GLsizeiptr',
    'float':         'GLfloat',
    'int':           'GLint',
}


# Element types not listed here fall back to ``%p`` at the call site.
ARRAY_ELEM_KIND = {
    'GLfloat':     'MESA_TRACE_ELEM_FLOAT',
    'GLclampf':    'MESA_TRACE_ELEM_FLOAT',
    'GLdouble':    'MESA_TRACE_ELEM_DOUBLE',
    'GLclampd':    'MESA_TRACE_ELEM_DOUBLE',
    'GLbyte':      'MESA_TRACE_ELEM_BYTE',
    'GLshort':     'MESA_TRACE_ELEM_SHORT',
    'GLint':       'MESA_TRACE_ELEM_INT',
    'GLsizei':     'MESA_TRACE_ELEM_INT',
    'GLfixed':     'MESA_TRACE_ELEM_INT',
    'GLclampx':    'MESA_TRACE_ELEM_INT',
    'GLubyte':     'MESA_TRACE_ELEM_UBYTE',
    'GLushort':    'MESA_TRACE_ELEM_USHORT',
    'GLuint':      'MESA_TRACE_ELEM_UINT',
    'GLhalfNV':    'MESA_TRACE_ELEM_HALF',
    'GLint64':     'MESA_TRACE_ELEM_INT64',
    'GLuint64':    'MESA_TRACE_ELEM_UINT64',
    'GLintptr':    'MESA_TRACE_ELEM_INTPTR',
    'GLsizeiptr':  'MESA_TRACE_ELEM_INTPTR',
}


def array_count_expr(p):
    if p.counter:
        scale = p.count_scale if p.count_scale else 1
        if scale == 1:
            return '(size_t){0}'.format(p.counter)
        return '(size_t){0} * {1}'.format(p.counter, scale)
    if p.count and p.count >= 2:
        return str(p.count)
    return None


def classify_param(p):
    """Return one of:
       ('scalar', spec, expr)        - printed inline
       ('array', kind, count_expr, name)
       ('opaque', spec, expr)        - printed as %p / hex fallback
    """
    opaque = ('opaque', '%p', '(void *){0}'.format(p.name))

    if not p.is_pointer():
        ts = TYPE_ALIASES.get(p.type_string().strip(), p.type_string().strip())
        if ts in TYPE_FORMAT:
            spec, expr = TYPE_FORMAT[ts]
            return ('scalar', spec, expr.format(p=p.name))
        return ('scalar', '0x%x', p.name)

    base = p.get_base_type_string()

    if base in ('GLchar', 'GLcharARB'):
        ts = p.type_string().lstrip()
        if ts.startswith('const'):
            return ('scalar', '%s',
                    '{p} ? (const char *){p} : "(null)"'.format(p=p.name))
        return opaque

    if p.is_output:
        return opaque

    elem = TYPE_ALIASES.get(base, base)
    count_expr = array_count_expr(p)
    if elem in ARRAY_ELEM_KIND and count_expr is not None:
        return ('array', ARRAY_ELEM_KIND[elem], count_expr, p.name)

    return opaque


class PrintCode(gl_XML.gl_print_base):
    def __init__(self):
        super(PrintCode, self).__init__()
        self.name = 'api_trace_c.py'
        self.license = license.bsd_license_template % (
            'Copyright (C) 2026 Christian Gmeiner', 'Christian Gmeiner')

    def printRealHeader(self):
        print(header)

    def printRealFooter(self):
        pass

    def print_wrapper(self, f):
        params = [p for p in f.parameters if not p.is_padding]
        classified = [classify_param(p) for p in params]

        specs = []
        args = []
        prelude = []
        for p, c in zip(params, classified):
            if c[0] == 'array':
                _, kind, count_expr, name = c
                buf_name = '{0}_buf'.format(name)
                prelude.append('char {buf}[{sz}];'
                               .format(buf=buf_name, sz=TRACE_ARRAY_BUFSZ))
                prelude.append(
                    '_mesa_trace_format_array({buf}, sizeof({buf}),'
                    ' {name}, {count}, {kind});'
                    .format(buf=buf_name, name=name,
                            count=count_expr, kind=kind))
                specs.append('%s')
                args.append(buf_name)
            else:
                _, spec, expr = c
                specs.append(spec)
                args.append(expr)

        ret = f.return_type
        param_string = f.get_parameter_string()
        called = f.get_called_parameter_string()

        print('static {rt} GLAPIENTRY'.format(rt=ret))
        print('_mesa_trace_{name}({ps})'.format(name=f.name, ps=param_string))
        print('{')
        print('   GET_CURRENT_CONTEXT(ctx);')
        for stmt in prelude:
            print('   {0}'.format(stmt))
        if args:
            fmt_body = 'gl{0}({1})\\n'.format(f.name, ', '.join(specs))
            print('   _mesa_debug(ctx, "{fmt}", {args});'
                  .format(fmt=fmt_body, args=', '.join(args)))
        else:
            print('   _mesa_debug(ctx, "gl{0}()\\n");'.format(f.name))
        call = 'CALL_{name}(ctx->Dispatch.RealPublished, ({ca}));'.format(
            name=f.name, ca=called if called else '')
        if ret != 'void':
            print('   return {call}'.format(call=call))
        else:
            print('   {call}'.format(call=call))
        print('}')
        print('')

    def print_install(self, functions):
        print('bool')
        print('_mesa_init_dispatch_trace(struct gl_context *ctx)')
        print('{')
        print('   struct _glapi_table *table = _mesa_alloc_dispatch_table(false);')
        print('   if (!table)')
        print('      return false;')
        print('')
        for f in functions:
            print('   SET_{name}(table, _mesa_trace_{name});'.format(name=f.name))
        print('')
        print('   ctx->Dispatch.Trace = table;')
        print('   return true;')
        print('}')

    def printBody(self, api):
        functions = list(api.functionIterateByOffset())
        for f in functions:
            self.print_wrapper(f)
        self.print_install(functions)


if __name__ == '__main__':
    apiexec.print_glapi_file(PrintCode())



# Copyright 2024 Intel Corporation
# SPDX-License-Identifier: MIT

import argparse
import sys
from math import pi

a = 'a'
b = 'b'
c = 'c'

lower_fsign = [
    (('fsign', a), ('bcsel', ('!flt', 0, a), +1.0,
                    ('bcsel', ('!flt', a, 0), -1.0, 0.0))),
    (('fceil', a), ('fneg', ('ffloor', ('fneg', a)))),

    # inot is free on and/or/xor sources but not dests. Apply De Morgan's.
    (('inot', ('iand(is_used_once)', ('inot', a), b)), ('ior', a, ('inot', b))),
    (('inot', ('ior(is_used_once)', ('inot', a), b)), ('iand', a, ('inot', b))),
    (('inot', ('ixor(is_used_once)', ('inot', a), b)), ('ixor', a, b)),
    (('inot', ('iand(is_used_once)', a, b)), ('ior', ('inot', a), ('inot', b))),
    (('inot', ('ior(is_used_once)', a, b)), ('iand', ('inot', a), ('inot', b))),
    (('inot', ('ixor(is_used_once)', a, b)), ('ixor', ('inot', a), b)),

    # Remove the zeroing. Down-conversion is free but extracts are not.
    (('u2f32', ('extract_u8', a, 0)), ('u2f32', ('u2u8', a))),
    (('u2f32', ('extract_u16', a, 0)), ('u2f32', ('u2u16', a))),
    (('i2f32', ('extract_i8', a, 0)), ('i2f32', ('i2i8', a))),
    (('i2f32', ('extract_i16', a, 0)), ('i2f32', ('i2i16', a))),

    (('pack_half_2x16_split', a, b),
     ('pack_32_2x16_split', ('f2f16', a), ('f2f16', b))),
]


lower_bool = [
    # Try to use conditional modifiers more
    (('ieq', ('iand(is_used_once)', a, b), b),
     ('ieq', ('iand', ('inot', a), b), 0)),
    (('ine', ('iand(is_used_once)', a, b), b),
     ('ine', ('iand', ('inot', a), b), 0)),

    # Clean up after 1-bit phi lowering
    (('b2b32', ('b2b1', 'a@32')), a),

    # Turn boolean ops into bitwise logic to simplify the backend
    (('ine', 'a@1', 'b@1'), ('ixor', a, b)),
    (('ieq', 'a@1', 'b@1'), ('ixor', ('inot', a), b)),
    (('bcsel', 'a', 'b@1', 'c@1'), ('ior', ('iand', a, b), ('iand', ('inot', a), c))),

    # We do not support 64-bit bcsel, so lower special
    ((f'b2i64', 'a@1'), ('pack_64_2x32_split', ('bcsel', a, 1, 0), 0)),
]

for T, sizes, one in [('f', [16, 32], 1.0),
                      ('i', [8, 16, 32], 1),
                      ('b', [32], -1)]:
    for sz in sizes:
        if T in ['f', 'i']:
            lower_bool.extend([
                ((f'{T}neg', (f'b2{T}{sz}', ('inot', 'a@1'))),
                 ('bcsel', a, 0, -one)),
                ((f'{T}neg', (f'b2{T}{sz}', 'a@1')), ('bcsel', a, -one, 0)),
            ])

        lower_bool.extend([
            ((f'b2{T}{sz}', ('inot', 'a@1')), ('bcsel', a, 0, one)),
            ((f'b2{T}{sz}', 'a@1'), ('bcsel', a, one, 0)),
        ])


lower_bfloat_ops = [
    (('bfmul', a, b), ('bfmul_mixed_intel', a, ('bf2f', b))),
    (('bffma', a, b, c), ('bffma_mixed_intel', ('bf2f', a), b, c)),
]


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    parser.add_argument('output')
    args = parser.parse_args()

    sys.path.insert(0, args.import_path)
    import nir_algebraic  # pylint: disable=import-error

    with open(args.output, 'w', encoding='utf-8') as f:
        f.write('#include "jay_private.h"')

        f.write(nir_algebraic.AlgebraicPass(
            "jay_nir_lower_fsign", lower_fsign).render())
        f.write(nir_algebraic.AlgebraicPass(
            "jay_nir_lower_bool", lower_bool).render())
        f.write(nir_algebraic.AlgebraicPass(
            "jay_nir_lower_bfloat_math", lower_bfloat_ops).render())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright © 2026 Intel Corporation
# SPDX-License-Identifier: MIT

import argparse
import sys

VALID_COMMON_VK_OPTIONS = {
    "force_vk_devicename",
    "force_vk_vendor",
    "vk_lower_terminate_to_discard",
    "vk_require_astc",
}

def declare_options(android_version):
    import drirc_gen

    B = drirc_gen.DrircBool
    I = drirc_gen.DrircInt
    F = drirc_gen.DrircFloat
    U64 = drirc_gen.DrircUint64
    E = drirc_gen.DrircEnum
    EV = drirc_gen.DrircEnumValue

    debug_options = [
        # Workaround subgroups
        I("anv_assume_full_subgroups", 0, 0, 32,
          "Allow assuming full subgroups requirement even when it's not specified explicitly and set the given size",
          c_name="assume_full_subgroups"),
        B("anv_assume_full_subgroups_with_barrier", False,
          "Assume full subgroups requirement for compute shaders that use control barriers",
          c_name="assume_full_subgroups_with_barrier"),
        B("anv_assume_full_subgroups_with_shared_memory", False,
          "Allow assuming full subgroups requirement for shaders using shared memory even when it's not specified explicitly",
          c_name="assume_full_subgroups_with_shared_memory"),
        B("anv_brw_disable_subgroup_size_control", False,
          "Disable EXT_subgroup_size_control support when using brw compiler.",
          c_name="disable_subgroup_size_control"),
        B("anv_large_workgroup_non_coherent_image_workaround", False,
          "Fixup image coherency qualifier for certain shaders.",
          c_name="large_workgroup_non_coherent_image_workaround"),

        # Workaround various compiler related
        B("anv_emulate_read_without_format", android_version >= 35,
          "Emulate shaderStorageImageReadWithoutFormat with shader conversions",
          c_name="read_without_format_emu"),
        B("anv_sample_mask_out_opengl_behaviour", False,
          "Ignore sample mask out when having single sampled target",
          c_name="sample_mask_out_opengl_behaviour"),
        B("anv_disable_link_time_optimization", False,
          "Disable linking of graphics pipeline shaders",
          c_name="disable_lto"),
        F("lower_depth_range_rate", 1.0, 0.0, 1.0,
          "Lower depth range for fixing misrendering issues due to z coordinate float point interpolation accuracy",
          c_name="lower_depth_range_rate"),
        B("force_indirect_descriptors", False,
          "Use an indirection to access buffer/image/texture/sampler handles",
          c_name="force_indirect_descriptors"),
        B("limit_trig_input_range", False,
          "Limit trig input range to [-2p : 2p] to improve sin/cos calculation precision on Intel",
          c_name="limit_trig_input_range"),
        B("fp64_workaround_enabled", False,
          "Use softpf64 when the shader uses float64, but the device doesn't support that type",
          c_name="fp64_emu"),
        B("no_16bit", False,
          "Disable 16-bit instructions",
          c_name="no_16bit"),
        I("shader_spilling_rate", 11, 0, 100,
          "Speed up shader compilation by increasing number of spilled registers after ra_allocate failure",
          c_name="shader_spilling_rate"),
        B("anv_fs_sampler_undef_derivatives_workaround", False,
          "Fixes samplers in fragment shaders computing undefined values for derivatives with lanes disabled by control flow",
          c_name="fs_sampler_undef_derivatives_workaround"),
        B("anv_slm_robust_vectorization", False,
          "Use robust vectorization for SLM accesses",
          c_name="slm_robust_vectorization"),
        B("anv_xe2_r11g11b10_atomic_swap_wa", True,
          "Enable workaround for apps using atomic swaps on R11G11B10 images",
          c_name="r11g11b10_atomic_swap_wa"),

        # Workaround various driver
        B("always_flush_cache", False,
          "Enable flushing GPU caches with each draw call", c_name="always_flush_cache"),
        B("anv_force_filter_addr_rounding", False,
          "Force min/mag filter address rounding to be enabled even for NEAREST sampling",
          c_name="force_filter_addr_rounding"),
        B("anv_disable_fcv", False,
          "Disable FCV optimization",
          c_name="disable_fcv"),
        B("anv_enable_buffer_comp", False,
          "Enable CCS on buffers where possible",
          c_name="enable_buffer_comp"),
        B("anv_external_memory_implicit_sync", False,
          "Implicit sync on external BOs",
          c_name="external_memory_implicit_sync"),
        B("anv_fake_nonlocal_memory", False,
          "Present host-visible device-local memory types as non device-local",
          c_name="fake_nonlocal_mem"),
        B("anv_upper_bound_descriptor_pool_sampler", False,
          "Overallocate samplers in descriptor pools to workaround app bug",
          c_name="upper_bound_desc_pool_sampler"),
        B("anv_disable_drm_ccs_modifiers", False,
          "Disable DRM CCS modifier usage",
          c_name="disable_xe2_ccs_modifiers"),
        B("custom_border_colors_without_format", android_version == 0,
          "Enable custom border colors without format",
          c_name="custom_border_colors_without_format"),
        B("intel_sampler_route_to_lsc", False,
          "Specific toggle to enable sampler route to LSC",
          c_name="sampler_route_to_lsc"),
        B("intel_storage_cache_policy_wt", False,
          "Enable write-through cache policy for storage buffers/images",
          c_name="storage_l1_wt"),
        B("intel_tbimr", True,
          "Enable TBIMR tiled rendering",
          c_name="tbimr"),
        B("intel_te_distribution", True,
          "Enable tesselation distribution",
          c_name="te_distribution"),
        B("intel_vf_distribution", True,
          "Enable geometry distribution",
          c_name="vf_distribution"),
        B("anv_write_lookup_maps_unconditionally", False,
          "Unconditionally write lookup maps for BLAS update operation",
          c_name="write_lookup_maps_unconditionally"),

        # Workaround command emission
        B("anv_barrier_post_untyped_clear_shader", False,
          "Insert pipeline barriers post clearing shader on untyped data",
          c_name="barrier_post_untyped_clear_shader"),
        B("anv_barrier_post_typed_clear_shader", False,
          "Insert pipeline barriers post clearing shader on typed data",
          c_name="barrier_post_typed_clear_shader"),
        B("intel_enable_wa_14018912822", False,
          "Workaround for using zero blend constants",
          c_name="wa_14018912822"),
        B("intel_enable_wa_14024015672_msaa", False,
          "Workaround for RHWO MSAA",
          c_name="wa_14024015672_msaa"),

        # Workaround command emission, shader specific
        B("force_vk_typed_barrier_after_dispatch_to_compute", False,
          "Insert a barrier for typed resources after dispatch of a shader for other compute shaders"),
        B("force_vk_untyped_barrier_after_dispatch_to_compute", False,
          "Insert a barrier for untyped resources after dispatch of a shader for other compute shaders"),
        B("force_vk_typed_barrier_after_dispatch_to_top", False,
          "Insert a barrier for typed resources after dispatch of a shader for any other shader"),
        B("force_vk_untyped_barrier_after_dispatch_to_top", False,
          "Insert a barrier for untyped resources after dispatch of a shader for any other shader"),
    ]

    perf_options = [
        B("adaptive_sync", True,
          "Adapt the monitor sync to the application performance (when possible)",
          c_name="adaptive_sync"),
        I("generated_indirect_threshold", 4, 0, 0x7fffffff,
          "Indirect threshold count above which we start generating commands",
          c_name="generated_indirect_threshold"),
        I("generated_indirect_ring_threshold", 100, 0, 0x7fffffff,
          "Indirect threshold count above which we start generating commands in a ring buffer",
          c_name="generated_indirect_ring_threshold"),
        I("query_clear_with_blorp_threshold", 6, 0, 0x7fffffff,
          "Query threshold count above which query buffers are cleared with blorp",
          c_name="query_clear_with_blorp_threshold"),
        I("query_copy_with_shader_threshold", 6, 0, 0x7fffffff,
          "Query threshold count above which query copies are executed with a shader",
          c_name="query_copy_with_shader_threshold"),

        B("anv_disable_push_constant_alloc", True,
          "Disable push constant space allocations",
          c_name="disable_push_const_alloc"),
        I("anv_binding_table_block_size",
          4096, 1024, 128 * 1024,
          "Binding table block allocation size (3DSTATE_BINDING_TABLE_POOL_ALLOC)",
          c_name="bt_block_size"),
        B("anv_promote_cbv_to_push_buffers", False,
          "Promote CBV 64bit pointers in push constant data to push buffers",
          c_name="promote_cbv_push_buffer"),
        B("anv_state_cache_perf_fix", False,
          "Whether COMMON_SLICE_CHICKEN3 bit13 should be programmed to enable BTP+BTI RCC keying",
          c_name="state_cache_perf_fix"),
        B("anv_vf_component_packing", True,
          "Vertex fetching component packing",
          c_name="vf_comp_packing"),
        I("anv_enable_opt_divergent_atomics", 0, 0, 3,
          "Enable fusion of divergent atomics (see brw_divergent_atomics_flags)",
          c_name="opt_divergent_atomics"),
        I("anv_enable_opt_divergent_atomics_compute_only", 0, 0, 3,
          "Enable fusion of divergent atomics for compute shaders only (see brw_divergent_atomics_flags)",
          c_name="opt_divergent_atomics_compute_only"),
        B("intel_force_compute_surface_prefetch", True,
          "Enable binding table surface prefteching for compute shaders",
          c_name="cs_surface_prefetch"),
        B("intel_force_sampler_prefetch", False,
          "Enable binding table sampler prefteching",
          c_name="sampler_prefetch"),

        B("force_guc_low_latency", False,
          "Enable low latency GuC strategy.",
          c_name="guc_low_latency"),

        E("anv_stack_ids", 512, 256, 2048,
          [EV(256,  "256 stackids"),
           EV(512,  "512 stackids"),
           EV(1024, "1024 stackids"),
           EV(2048, "2048 stackids")],
          "Control the number stackIDs (i.e. number of unique rays in the RT subsytem)",
          c_name="stack_ids"),
        E("anv_rt_dispatch_timeout", 512, 64, 4096,
          [EV(64,    "64 clocks"),
           EV(128,   "128 clocks"),
           EV(192,   "192 clocks"),
           EV(256,   "256 clocks"),
           EV(384,   "384 clocks"),
           EV(512,   "512 clocks"),
           EV(640,   "640 clocks"),
           EV(768,   "768 clocks"),
           EV(896,   "896 clocks"),
           EV(1024,  "1024 clocks"),
           EV(1152,  "1152 clocks"),
           EV(1280,  "1280 clocks"),
           EV(1408,  "1408 clocks"),
           EV(1536,  "1536 clocks"),
           EV(1664,  "1664 clocks"),
           EV(1792,  "1792 clocks"),
           EV(1920,  "1920 clocks"),
           EV(2048,  "2048 clocks"),
           EV(4096,  "4096 clocks")],
          "Force BTD child dispatches if dispatches do not happen naturally for number of clocks equal to the programmed timeout counter",
          c_name="rt_dispatch_timeout"),
    ]

    feature_options = [
        B("fake_sparse", False,
          "Advertise support for sparse binding of textures regardless of real support",
          c_name="fake_sparse"),
        B("anv_enable_scratch_page", True,
          "Disables surface padding and suppresses all page faults, drops writes and returns zeros on reads.",
          c_name="scratch_page"),
        B("anv_enable_fully_covered", False,
          "Enable fullyCoveredFragmentShaderInputVariable (Alchemist and newer only).",
          c_name="fully_covered"),
        B("anv_fake_image_compression_control_xe2_plus", android_version >= 37,
          "Enable VK_EXT_image_compression_control with no actual effect",
          c_name= "fake_image_compression_control_xe2_plus"),
        B("compression_control_enabled", android_version >= 37,
          "Enable VK_EXT_image_compression_control support",
          c_name="compression_control_enabled"),
    ]

    misc_options = []

    drirc_gen.add_common_vk_options(debug_options, feature_options, misc_options,
                                    valid_options=VALID_COMMON_VK_OPTIONS,
                                    defaults={"vk_require_astc": android_version >= 34})
    drirc_gen.add_common_vk_wsi_options(debug_options, perf_options)

    return [drirc_gen.DrircSection("Debugging",   debug_options,   c_name="debug"),
            drirc_gen.DrircSection("Features",    feature_options, c_name="features"),
            drirc_gen.DrircSection("Performance", perf_options,    c_name="perf")]

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    parser.add_argument('--drirc-src', required=True)
    parser.add_argument('--drirc-hdr', required=True)
    parser.add_argument('--android-ver', type=int, default=0, required=False)
    parser.add_argument('--validate', required=True)
    args = parser.parse_args()

    sys.path.insert(0, args.import_path)
    import drirc_gen

    options = declare_options(args.android_ver)

    drirc_gen.drirc_validate([args.validate], options)

    drirc_gen.drirc_generate(args.drirc_src, args.drirc_hdr, "anv", options)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright © 2025 Intel Corporation
# SPDX-License-Identifier: MIT

import argparse
import copy
import ctypes
import enum
import fcntl
import json
import pathlib
import os
import re
import sys

from gen_opcodes import OPCODES, Props


def c_bool(value):
    return 'true' if value else 'false'


def opcode_has_dst(opcode):
    return not bool(opcode.props & Props.NO_DST)


H_TEMPLATE = """\
/*
 * Copyright © 2026 Intel Corporation
 * SPDX-License-Identifier: MIT
 *
 * Autogenerated file, do not edit!
 */

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus

<%def name="comma_join(l)">${', '.join([str(i) for i in l])}</%def>

struct gen_encoding_${info.name.lower()} {
   static constexpr gen_encoding_type TYPE = GEN_ENCODING_${info.name.upper()};

   /* Instruction fields */
% for (i, rs) in info.fields.items():
   static constexpr auto ${i} = gen_ranges<${len(rs)}>({
% for r in rs:
      gen_range { ${comma_join(r)} },
% endfor
   });
% endfor

   /* Compact instruction fields */
% for (i, rs) in info.compact_fields.items():
   static constexpr auto C_${i} = gen_ranges<${len(rs)}>({
% for r in rs:
      gen_range { ${comma_join(r)} },
% endfor
   });
% endfor

   /* Instruction sub fields */
% for (i, rs) in info.sub_fields.items():
   static constexpr auto ${i} = gen_sub_ranges<${len(rs)}>({
% for r in rs:
      gen_sub_range { ${comma_join(r)} },
% endfor
   });
% endfor

   static constexpr std::array<gen_inst_description, 128> gen_to_description =
   []() constexpr {
      std::array<gen_inst_description, 128> r;
% for (op, hw) in info.gen_op_to_hw:
      r[GEN_OP_${op.enum_name}] = gen_inst_description(GEN_OP_${op.enum_name}, ${hw}, ${op.format}, ${c_bool(opcode_has_dst(op))});
% endfor
      return r;
   }();

   static constexpr std::array<gen_inst_description, 128> hw_to_description =
   []() constexpr {
      std::array<gen_inst_description, 128> r;
% for (op, hw) in info.gen_op_to_hw:
      r[${hw}] = gen_inst_description(GEN_OP_${op.enum_name}, ${hw}, ${op.format}, ${c_bool(opcode_has_dst(op))});
% endfor
      return r;
   }();
};

#endif

#ifdef __cplusplus
}
#endif
"""


class InstructionInfo:

    def __init__(self, raw_info, name):
        self.name = name
        self.imported = set()
        self.target = self.__init_with_inheritance(raw_info, name)
        self.__process_ranges()
        assert name in self.imported

    def __init_with_inheritance(self, raw_info, name):
        assert name in raw_info
        import_name = raw_info[name].get('import')
        if import_name:
            assert import_name not in self.imported
            self.imported.add(import_name)
            base = self.__init_with_inheritance(raw_info, import_name)
        else:
            base = dict()

        def find_dicts(d):
            return set({ k for k, v in d.items() if isinstance(v, dict) })

        new_target = copy.deepcopy(raw_info[name])

        # Drop all comment items. (object key names that start with a '#')
        for d in find_dicts(new_target):
            new_target[d] = { k: v for k, v
                              in new_target[d].items()
                              if not k.startswith('#') }

        if import_name:
            needed = find_dicts(new_target) - find_dicts(base)
            base.update({ k: {} for k in needed })
            needed = find_dicts(base) - find_dicts(new_target)
            new_target.update({ k: {} for k in needed })
            # Merge imported fields, but prefer current version
            dicts = find_dicts(new_target)
            for d in dicts:
                old_only = { k: v for k, v in base[d].items()
                             if k not in new_target[d] }
                assert len(set(old_only.keys()).intersection(set(new_target[d].keys()))) == 0
                new_target[d].update(old_only)

        required_infos = ('fields', 'sub-fields', 'compact-fields')
        for info in required_infos:
            if info not in new_target:
                new_target[info] = {}

        # Drop all items that are None (null in json)
        for d in find_dicts(new_target):
            new_target[d] = { k: v for k, v in new_target[d].items()
                              if v is not None }

        self.imported.add(name)
        return new_target

    def __process_ranges(self):
        """Process ranges in self.target

        On input, the fields are lists of integers. On output, they
        are list of tuples of length 2 where the first item is the
        high bit number of the range and the second item is the low
        bit number of the range.

        Additionally, any adjacent ranges that happen to have bit
        ranges that can be combined are collapsed into a single range.

        """

        def collapse_ranges(ranges):
            """Generator that merges compatible ranges."""

            it = iter(ranges)
            r = next(it)
            for n in it:
                if (r[1] - 1) == n[0]:
                    r = (r[0], n[1])
                else:
                    yield r
                    r = n
            yield r

        def pair_up(v):
            if v is None:
                return None
            assert (len(v) % 2) == 0
            ranges = [(v[i], v[i + 1]) for i in range(0, len(v), 2)]
            return list(collapse_ranges(ranges));

        for field_type in ('fields', 'sub-fields', 'compact-fields'):
            d = self.target.get(field_type, dict())
            d = { k: pair_up(v) for k, v in d.items() }
            self.target[field_type] = d

    def __getattr__(self, name):
        dashed_name = name.replace('_', '-')
        if dashed_name in self.target:
            return self.target[dashed_name]
        raise AttributeError

    @property
    def gen_op_to_hw(self):
        encoding = self.name.lower()
        result = []
        for op in OPCODES:
            hw = op.hw_opcode[encoding]
            if hw is not None:
                result.append((op, hw))
        return result

    def for_mako(self):
        return self.target

class GenInstructionSources:

    def __init__(self, args):
        self.args = args
        self.raw_info = {}
        for in_json in self.args.json:
            in_json = pathlib.Path(in_json)
            assert in_json.is_file()
            self.outdir = pathlib.Path(args.output_dir)
            assert self.outdir.is_dir()
            with open(in_json, 'r',  encoding='utf8') as json_file:
                new_info = json.load(json_file)
                # Make sure there is no overlap in json input
                assert all({k not in self.raw_info for k in new_info.keys()})
                self.raw_info.update(new_info)

        import mako.template
        self.h_template = mako.template.Template(H_TEMPLATE)

    def gen(self):
        gen_count = 0
        if self.args.cpp:
            gen_count += self.gen_cpp_headers()
        if gen_count == 0:
            print("info: No sources specified to be generated")
        return gen_count > 0

    def gen_cpp_header(self, name):
        info = InstructionInfo(self.raw_info, name)
        template_input = {
            "info": info,
            "c_bool": c_bool,
            "opcode_has_dst": opcode_has_dst,
        }
        filename = self.outdir / f'gen_info_{name.lower()}.h'
        with open(filename, 'w', encoding='utf8') as h:
            try:
                h.write(self.h_template.render(**template_input))
            except:
                import mako.exceptions
                print(mako.exceptions.text_error_template().render())
                raise
        return True

    def gen_cpp_headers(self):
        count = 0
        names = list(self.raw_info.keys())
        for name in names:
            if self.gen_cpp_header(name):
                count += 1
        return count

class GenInstructionsApp:

    def __init__(self):
        self.parse_args()
        self.gen_sources = GenInstructionSources(self.args)
        sys.exit(0 if self.gen_sources.gen() else 1)

    def parse_args(self):
        p = argparse.ArgumentParser()
        p.add_argument("-v", "--verbose", action="count", default=0,
                       help="Enable verbose output")
        p.add_argument("--json", action="append", default=[],
                       help="Input json describing the target")
        p.add_argument("--output-dir", required=True,
                       help="Output directory for generated files")
        p.add_argument("--cpp", action="store_true", default=False,
                       help="Generate C++ headers")

        a = p.parse_args()

        self.args = a
        self.verbose = a.verbose


if __name__ == "__main__":
    GenInstructionsApp()



# Copyright © 2026 Intel Corporation
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
import enum


@enum.unique
class Props(enum.IntFlag):
    NONE = 0
    HAS_JIP = 1 << 0
    HAS_UIP = 1 << 1
    BRANCH_CTRL = 1 << 2
    NO_DST = 1 << 3


HwOpcodeValue = int | None
HwOpcodeInput = int | dict[str, HwOpcodeValue]
HwOpcode = dict[str, HwOpcodeValue]


@dataclass(frozen=True)
class Opcode:
    name: str
    enum_name: str
    format: str
    hw_opcode: HwOpcode
    num_srcs: int
    props: Props = Props.NONE


def normalize_hw_opcode(hw_opcode: HwOpcodeInput) -> HwOpcode:
    if isinstance(hw_opcode, int):
        return {"pre_xe": hw_opcode, "xe": hw_opcode, "xe2": hw_opcode,
                "xe3p": hw_opcode}

    assert hw_opcode.keys() <= {"pre_xe", "xe", "xe2", "xe3p"}
    assert all(value is None or isinstance(value, int)
               for value in hw_opcode.values())
    pre_xe = hw_opcode.get("pre_xe")
    xe = hw_opcode.get("xe", pre_xe)
    xe2 = hw_opcode.get("xe2", xe)
    xe3p = hw_opcode.get("xe3p", xe2)
    return {
        "pre_xe": pre_xe,
        "xe": xe,
        "xe2": xe2,
        "xe3p": xe3p,
    }


_opcodes: list[Opcode] = []


def op(name: str, format: str, hw_opcode: HwOpcodeInput,
       props: Props = Props.NONE, *, num_srcs: int) -> None:
    _opcodes.append(Opcode(
        name=name,
        enum_name=name.upper(),
        format=format,
        hw_opcode=normalize_hw_opcode(hw_opcode),
        props=props,
        num_srcs=num_srcs,
    ))


def basic1(name: str, hw_opcode: HwOpcodeInput,
           props: Props = Props.NONE) -> None:
    op(name, 'GEN_FORMAT_BASIC_ONE_SRC', hw_opcode, props, num_srcs=1)


def basic2(name: str, hw_opcode: HwOpcodeInput,
           props: Props = Props.NONE) -> None:
    op(name, 'GEN_FORMAT_BASIC_TWO_SRC', hw_opcode, props, num_srcs=2)


def basic3(name: str, hw_opcode: HwOpcodeInput,
           props: Props = Props.NONE) -> None:
    op(name, 'GEN_FORMAT_BASIC_THREE_SRC', hw_opcode, props, num_srcs=3)


def branch1(name: str, hw_opcode: HwOpcodeInput,
            props: Props = Props.NONE) -> None:
    branch_props = Props.NO_DST
    if name != 'join':
        branch_props |= Props.BRANCH_CTRL
    op(name, 'GEN_FORMAT_BRANCH_ONE_SRC', hw_opcode, branch_props | props,
       num_srcs=1)


def branch2(name: str, hw_opcode: HwOpcodeInput,
            props: Props = Props.NONE) -> None:
    op(name, 'GEN_FORMAT_BRANCH_TWO_SRC', hw_opcode,
       Props.BRANCH_CTRL | Props.NO_DST | props, num_srcs=2)


op('illegal', 'GEN_FORMAT_ILLEGAL', 0, Props.NO_DST, num_srcs=0)

basic1('bfrev', {"pre_xe": 23, "xe": 119})
basic1('cbit',  77)
basic1('fbh',   75)
basic1('fbl',   76)
basic1('frc',   67)
basic1('lzd',   74)
basic1('mov',   {"pre_xe": 1, "xe": 97})
basic1('not',   {"pre_xe": 4, "xe": 100})
basic1('rndd',  69)
basic1('rnde',  70)
basic1('rndu',  68)
basic1('rndz',  71)
basic1('sync',  {"xe": 1}, Props.NO_DST)
basic1('wait',  48)

basic2('add',   64)
basic2('addc',  78)
basic2('and',   {"pre_xe": 5, "xe": 101})
basic2('asr',   {"pre_xe": 12, "xe": 108})
basic2('avg',   66)
basic2('bfi1',  {"pre_xe": 25, "xe": 121})
basic2('cmp',   {"pre_xe": 16, "xe": 112})
basic2('cmpn',  {"pre_xe": 17, "xe": 113})
basic2('dp2',   {"pre_xe": 87, "xe": None})
basic2('dp3',   {"pre_xe": 86, "xe": None})
basic2('dp4',   {"pre_xe": 84, "xe": None})
basic2('dph',   {"pre_xe": 85, "xe": None})
basic2('line',  {"pre_xe": 89, "xe": None})
basic2('mac',   {"pre_xe": 72, "xe3p": None})
basic2('mach',  {"pre_xe": 73, "xe3p": None})
basic2('macl',  {"xe2": 83, "xe3p": None})
basic2('math',  56)
basic2('movi',  {"pre_xe": 3, "xe": 99})
basic2('mul',   65)
basic2('mullh', {"xe3p": 95})
basic2('or',    {"pre_xe": 6, "xe": 102})
basic2('pln',   {"pre_xe": 90, "xe": None})
basic2('rol',   {"pre_xe": 15, "xe": 111})
basic2('ror',   {"pre_xe": 14, "xe": 110})
basic2('sel',   {"pre_xe": 2, "xe": 98})
basic2('shl',   {"pre_xe": 9, "xe": 105})
basic2('shr',   {"pre_xe": 8, "xe": 104})
basic2('smov',  {"pre_xe": 10, "xe": 106}, Props.NO_DST)
basic2('srnd',  {"xe2": 84})
basic2('subb',  79)
basic2('xor',   {"pre_xe": 7, "xe": 103})

basic3('add3', {"xe": 82})
basic3('bfe',  {"pre_xe": 24, "xe": 120})
basic3('bfi2', {"pre_xe": 26, "xe": 122})
basic3('bfn',  {"xe": 107})
basic3('csel', {"pre_xe": 18, "xe": 114})
basic3('dp4a', {"xe": 88})
basic3('lrp',  {"pre_xe": 92, "xe": None})
basic3('mad',  91)
basic3('madm', 93)

op('dpas', 'GEN_FORMAT_DPAS_THREE_SRC', {"xe": 89}, num_srcs=3)

op('send',   'GEN_FORMAT_SEND', 49, num_srcs=1)
op('sendc',  'GEN_FORMAT_SEND', 50, num_srcs=1)
op('sends',  'GEN_FORMAT_SEND', {"pre_xe": 51, "xe": None}, num_srcs=1)
op('sendsc', 'GEN_FORMAT_SEND', {"pre_xe": 52, "xe": None}, num_srcs=1)

branch1('brd',   33)
branch1('call',  44)
branch1('calla', 43)
branch1('endif', 37, Props.HAS_JIP)
branch1('jmpi',  32)
branch1('join',  47, Props.HAS_JIP)
branch1('ret',   45)
branch1('while', 39, Props.HAS_JIP)

branch2('brc',      35)
branch2('break',    40, Props.HAS_JIP | Props.HAS_UIP)
branch2('continue', 41, Props.HAS_JIP | Props.HAS_UIP)
branch2('else',     36, Props.HAS_JIP | Props.HAS_UIP)
branch2('goto',     46, Props.HAS_JIP | Props.HAS_UIP)
branch2('halt',     42, Props.HAS_JIP | Props.HAS_UIP)
branch2('if',       34, Props.HAS_JIP | Props.HAS_UIP)

op('nop', 'GEN_FORMAT_NOP', {"pre_xe": 126, "xe": 96}, Props.NO_DST, num_srcs=0)

OPCODES = tuple(_opcodes)
OPCODES_BY_ENUM_NAME = {op.enum_name: op for op in OPCODES}


del op
del _opcodes
